# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from logging.handlers import TimedRotatingFileHandler
from iconcommons.logger.utils import suffix as rotate_suffix, extMatch as rotate_extMatch


class IconTimeRotatingFileHandler(TimedRotatingFileHandler):
    def __init__(self, filename, when='h', interval=1, backupCount=0,
                 encoding=None, delay=False, utc=False, atTime=None):
        super().__init__(filename, when, interval, backupCount, encoding, delay, utc, atTime)

    def getFilesToDelete(self):
        """
        Determine the files to delete when rolling over.

        More specific than the earlier method, which just used glob.glob().
        """
        dirName, baseName = os.path.split(self.baseFilename)
        fileNames = os.listdir(dirName)
        result = []
        prefix = baseName + "."
        plen = len(prefix)
        for fileName in fileNames:
            if fileName[:plen] == prefix:
                suffix = fileName[plen:]
                if rotate_extMatch.match(suffix):
                    result.append(os.path.join(dirName, fileName))
        if len(result) < self.backupCount:
            result = []
        else:
            result.sort()
            result = result[:len(result) - self.backupCount]
        return result

    def doRollover(self):
        """
        do a rollover; in this case, a date/time stamp is appended to the filename
        when the rollover happens.  However, you want the file to be named for the
        start of the interval, not the current time.  If there is a backup count,
        then we have to get a list of matching filenames, sort them and remove
        the one with the oldest suffix.
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # get the time that this sequence started at and make it a TimeTuple
        currentTime = int(time.time())
        dstNow = time.localtime(currentTime)[-1]
        t = self.rolloverAt - self.interval
        if self.utc:
            timeTuple = time.gmtime(t)
        else:
            timeTuple = time.localtime(t)
            dstThen = timeTuple[-1]
            if dstNow != dstThen:
                if dstNow:
                    addend = 3600
                else:
                    addend = -3600
                timeTuple = time.localtime(t + addend)
        dfn = self.rotation_filename(self.baseFilename + "." +
                                     time.strftime(rotate_suffix, timeTuple))
        if os.path.exists(dfn):
            os.remove(dfn)
        self.rotate(self.baseFilename, dfn)
        if self.backupCount > 0:
            for s in self.getFilesToDelete():
                os.remove(s)
        if not self.delay:
            self.stream = self._open()
        newRolloverAt = self.computeRollover(currentTime)
        while newRolloverAt <= currentTime:
            newRolloverAt = newRolloverAt + self.interval
        #If DST changes and midnight or weekly rollover, adjust for this.
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            dstAtRollover = time.localtime(newRolloverAt)[-1]
            if dstNow != dstAtRollover:
                if not dstNow:  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:           # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                newRolloverAt += addend
        self.rolloverAt = newRolloverAt



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from enum import Flag
from logging import Logger as builtinLogger, FileHandler, Formatter, Handler
from typing import Union

from .icon_period_and_bytes_file_handler import IconPeriodAndBytesFileHandler
from .icon_rotating_file_handler import IconRotatingFileHandler
from .icon_time_rotating_file_handler import IconTimeRotatingFileHandler


class OutputType(Flag):
    NONE = 0
    CONSOLE = 1
    FILE = 2


class RotateType(Flag):
    NONE = 0
    PERIOD = 1
    BYTES = 2
    BOTH = PERIOD | BYTES


class RotateConfig:
    def __init__(self,
                 rotate_type: 'RotateType',
                 period: str,
                 interval: int,
                 max_bytes: int,
                 backup_count: int):
        self.rotate_type: 'RotateType' = rotate_type
        self.period: str = period
        self.interval: int = interval
        self.max_bytes: int = max_bytes
        self.backup_count: int = backup_count

    @classmethod
    def from_dict(cls, src_config: dict):
        config: dict = src_config.get('rotate')
        if config is None:
            return

        rotate_type: 'RotateType' = RotateType.NONE

        rotate_types: str = config.get('type')
        if rotate_types:
            rotates = rotate_types.split('|')
            for rotate in rotates:
                rotate_type |= RotateType[rotate.upper()]

        period: str = cls._convert_period(config.get('period', "daily"))
        interval: int = config.get('interval')
        max_bytes: int = config.get('maxBytes')
        backup_count: int = config.get("backupCount")

        return RotateConfig(rotate_type, period, interval, max_bytes, backup_count)

    @classmethod
    def _convert_period(cls, value: str):
        if value == 'daily':
            return 'D'
        elif value == 'weekly':
            return 'W'
        elif value == 'hourly':
            return 'H'
        elif value == 'minutely':
            return 'M'
        elif value == 'secondly':
            return 'S'
        else:
            return 'D'


class LogConfig:
    DEFAULT_FORMAT = "%(asctime)s %(process)d %(thread)d %(levelname)s %(filename)s(%(lineno)d) %(message)s"

    def __init__(self,
                 name: str,
                 level: str,
                 file_path: str,
                 fmt: str,
                 output_type: 'OutputType',
                 rotate_config: 'RotateConfig'):

        self.name: str = name
        self.level: str = level
        self.file_path: str = file_path
        self.fmt: str = fmt
        self.output_type: 'OutputType' = output_type
        self.rotate_config: 'RotateConfig' = rotate_config

    @classmethod
    def from_dict(cls, src_config: dict):
        config: dict = src_config.get('log')
        if config is None:
            return

        name: str = config.get('name', "ICONLogger")
        level: str = config.get('level', 'info').upper()
        file_path: str = config.get('filePath', "")
        fmt: str = config.get('format', cls.DEFAULT_FORMAT)
        output_type: 'OutputType' = OutputType.NONE

        output_types: str = config.get('outputType')
        if output_types:
            outputs = output_types.split('|')
            for output in outputs:
                output_type |= OutputType[output.upper()]

        rotate_config: 'RotateConfig' = RotateConfig.from_dict(config)
        return LogConfig(name, level, file_path, fmt, output_type, rotate_config)


class IconLoggerUtil(object):
    _formatter: 'Formatter' = None

    @classmethod
    def apply_config(cls, logger: 'builtinLogger', config: dict) -> None:
        log_config: 'LogConfig' = LogConfig.from_dict(config)

        logger.handlers.clear()
        logger.name = log_config.name
        logger.setLevel(log_config.level)
        cls._apply_config(logger, log_config)

    @classmethod
    def print_config(cls, logger: 'builtinLogger', config: dict):
        logger.info(f'====================LOG CONFIG START====================')
        cls._view_config_info(logger, config, "CONFIG")
        logger.info(f'====================LOG CONFIG END======================')

    @classmethod
    def _view_config_info(cls, logger: 'builtinLogger', conf: dict, prefix: str):
        for key, value in conf.items():
            if not isinstance(value, dict):
                tmp_prefix = '{}.{}'.format(prefix, key)
                logger.info(f'[{tmp_prefix}] > {value}')
            else:
                tmp_prefix = '{}.{}'.format(prefix, key)
                cls._view_config_info(logger, value, tmp_prefix)

    @classmethod
    def _apply_config(cls, logger: 'builtinLogger', log_config: 'LogConfig'):
        cls._formatter = Formatter(log_config.fmt)

        if cls._is_flag_on(log_config.output_type, OutputType.CONSOLE):
            handler = logging.StreamHandler()
            handler.setFormatter(cls._formatter)
            logger.addHandler(handler)

        if cls._is_flag_on(log_config.output_type, OutputType.FILE):
            cls._ensure_dir(log_config.file_path)

            if log_config.rotate_config is None:
                handler = cls.make_file_handler(log_config.file_path, cls._formatter)
                logger.addHandler(handler)
            else:
                rotate_type: 'Flag' = log_config.rotate_config.rotate_type
                if cls._is_flag_on(rotate_type, RotateType.BOTH):
                    handler = cls.make_period_and_bytes_file_handler(log_config.file_path,
                                                                     log_config.rotate_config.period,
                                                                     log_config.rotate_config.interval,
                                                                     log_config.rotate_config.max_bytes,
                                                                     log_config.rotate_config.backup_count)
                    logger.addHandler(handler)
                elif cls._is_flag_on(rotate_type, RotateType.PERIOD):
                    handler = cls.make_period_file_handler(log_config.file_path,
                                                           log_config.rotate_config.period,
                                                           log_config.rotate_config.interval,
                                                           log_config.rotate_config.backup_count)
                    logger.addHandler(handler)
                elif cls._is_flag_on(rotate_type, RotateType.BYTES):
                    handler = cls.make_bytes_file_handler(log_config.file_path,
                                                          log_config.rotate_config.max_bytes,
                                                          log_config.rotate_config.backup_count)
                    logger.addHandler(handler)

    @classmethod
    def _is_flag_on(cls, src_flag: 'Flag', dest_flag: 'Flag') -> bool:
        return src_flag & dest_flag == dest_flag

    @classmethod
    def _make_exc_log_path(cls, src_path: str) -> str:
        if src_path.endswith(".log"):
            converted = src_path[:-4] + "_exc.log"
        else:
            converted = src_path + "_exc"
        return converted

    @classmethod
    def _ensure_dir(cls, file_path: str):
        directory = os.path.dirname(file_path)
        if not os.path.exists(directory):
            os.makedirs(directory)

    @classmethod
    def make_log_msg(cls, tag: str, msg: Union[str, BaseException]):
        return f'{tag} {msg}'

    @classmethod
    def make_file_handler(cls,
                          file_path: str,
                          formatter: 'Formatter') -> 'Handler':
        handler = FileHandler(file_path, 'a')
        handler.setFormatter(formatter)
        return handler

    @classmethod
    def make_period_and_bytes_file_handler(cls,
                                           file_path: str,
                                           when: str,
                                           interval: int,
                                           max_bytes: int,
                                           backup_count: int) -> 'Handler':
        handler = IconPeriodAndBytesFileHandler(file_path,
                                                maxBytes=max_bytes,
                                                when=when,
                                                interval=interval,
                                                backupCount=backup_count)
        handler.setFormatter(cls._formatter)
        return handler

    @classmethod
    def make_period_file_handler(cls,
                                 file_path: str,
                                 when: str,
                                 interval: int,
                                 backup_count: int) -> 'Handler':
        handler = IconTimeRotatingFileHandler(file_path,
                                              when=when,
                                              interval=interval,
                                              backupCount=backup_count)
        handler.setFormatter(cls._formatter)
        return handler

    @classmethod
    def make_bytes_file_handler(cls,
                                file_path: str,
                                max_bytes: int,
                                backup_count: int) -> 'Handler':
        handler = IconRotatingFileHandler(file_path,
                                          maxBytes=max_bytes,
                                          backupCount=backup_count)
        handler.setFormatter(cls._formatter)
        return handler


icon_logger = logging.Logger("ICONLogger")



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import types
from logging.handlers import BaseRotatingHandler
from stat import ST_MTIME

from .icon_rotating_file_handler import IconRotatingFileHandler
from .icon_time_rotating_file_handler import IconTimeRotatingFileHandler


class IconPeriodAndBytesFileHandler(BaseRotatingHandler):
    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0,
                 encoding=None, delay=0, when='h', interval=1, utc=False):
        super().__init__(filename, mode, encoding, delay)

        '''
        In fact the difference between adding methods dynamically at run time and your example is huge:

        in your case, you just attach a function to an object, you can call it of course but it is unbound, 
        it has no relation with the object itself (ie. you cannot use self inside the function)
        when added with MethodType, you create a bound method and it behaves like a normal Python method for the object, 
        you have to take the object it belongs to as first argument 
        (it is normally called self) and you can access it inside the function
        This example shows the difference:
        
        def func(obj):
          print 'I am called from', obj
        class A:
          pass
        a=A()
        a.func=func
        a.func()
        This fails with a TypeError: func() takes exactly 1 argument (0 given), whereas this code works as expected:
        
        import types
        a.func = types.MethodType(func, a, A)
        a.func()
        '''

        self.computeRollover = types.MethodType(IconTimeRotatingFileHandler.computeRollover, self)
        self.doRollover_bytes = types.MethodType(IconRotatingFileHandler.doRollover, self)
        self.getFilesToDelete = types.MethodType(IconRotatingFileHandler.getFilesToDelete, self)
        self.shouldRollover_period = types.MethodType(IconTimeRotatingFileHandler.shouldRollover, self)
        self.shouldRollover_bytes = types.MethodType(IconRotatingFileHandler.shouldRollover, self)

        # period
        self.maxBytes = maxBytes
        self.backupCount = backupCount
        self.when = when.upper()
        self.interval = interval
        self.utc = utc

        if self.when == 'S':
            self.interval = 1 # one second
        elif self.when == 'M':
            self.interval = 60 # one minute
        elif self.when == 'H':
            self.interval = 60 * 60 # one hour
        elif self.when == 'D' or self.when == 'MIDNIGHT':
            self.interval = 60 * 60 * 24 # one day
        elif self.when.startswith('W'):
            self.interval = 60 * 60 * 24 * 7 # one week
            if len(self.when) != 2:
                raise ValueError("You must specify a day for weekly rollover from 0 to 6 (0 is Monday): %s" % self.when)
            if self.when[1] < '0' or self.when[1] > '6':
                raise ValueError("Invalid day specified for weekly rollover: %s" % self.when)
            self.dayOfWeek = int(self.when[1])
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)

        self.interval = self.interval * interval # multiply by units requested
        # The following line added because the filename passed in could be a
        # path object (see Issue #27493), but self.baseFilename will be a string
        filename = self.baseFilename
        if os.path.exists(filename):
            t = os.stat(filename)[ST_MTIME]
        else:
            t = int(time.time())
        self.rolloverAt = self.computeRollover(t)

        # bytes
        self.rotator = types.MethodType(IconRotatingFileHandler.custom_rotator, self)
        self.logger_index = 0

    def doRollover(self):
        # custom bytes + period

        # get from logging.handlers.TimedRotatingFileHandler.doRollover()
        current_time = int(time.time())
        dst_now = time.localtime(current_time)[-1]

        new_rollover_at = self.computeRollover(current_time)
        while new_rollover_at <= current_time:
            new_rollover_at = new_rollover_at + self.interval

        # If DST changes and midnight or weekly rollover, adjust for this.
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            dst_at_rollover = time.localtime(new_rollover_at)[-1]
            if dst_now != dst_at_rollover:
                if not dst_now:  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:  # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                new_rollover_at += addend
        self.rolloverAt = new_rollover_at

        # support multify file
        return self.doRollover_bytes()

    def shouldRollover(self, record):
        return self.shouldRollover_period(record) or self.shouldRollover_bytes(record)



# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from logging import DEBUG, INFO, WARNING, ERROR, currentframe

from .icon_logger_util import IconLoggerUtil, icon_logger

# This code is mainly copied from the python logging module, with minor modifications

# _srcfile is used when walking the stack to check when we've got the first
# caller stack frame.
#

if hasattr(sys, 'frozen'): #support for py2exe
    _srcfile = "logging%s__init__%s" % (os.sep, __file__[-4:])
elif __file__[-4:].lower() in ['.pyc', '.pyo']:
    _srcfile = __file__[:-4] + '.py'
else:
    _srcfile = __file__
_srcfile = os.path.normcase(_srcfile)


class Logger(object):
    # for backward compatibility

    @classmethod
    def load_config(cls, config: dict):
        IconLoggerUtil.apply_config(icon_logger, config)

    @classmethod
    def print_config(cls, config: dict, tag: str):
        IconLoggerUtil.print_config(icon_logger, config)

    @classmethod
    def debug(cls, msg: str, tag: str = "LOG"):
        if icon_logger.isEnabledFor(DEBUG):
            cls._log(DEBUG, IconLoggerUtil.make_log_msg(tag, msg))

    @classmethod
    def info(cls, msg: str, tag: str = "LOG"):
        if icon_logger.isEnabledFor(INFO):
            cls._log(INFO, IconLoggerUtil.make_log_msg(tag, msg))

    @classmethod
    def warning(cls, msg: str, tag: str = "LOG"):
        if icon_logger.isEnabledFor(WARNING):
            cls._log(WARNING, IconLoggerUtil.make_log_msg(tag, msg))

    @classmethod
    def error(cls, msg: str, tag: str = "LOG"):
        if icon_logger.isEnabledFor(ERROR):
            cls._log(ERROR, IconLoggerUtil.make_log_msg(tag, msg))

    @classmethod
    def exception(cls, msg: str, tag: str = "LOG"):
        if icon_logger.isEnabledFor(ERROR):
            cls._log(ERROR, IconLoggerUtil.make_log_msg(tag, msg), exc_info=True)

    @classmethod
    def _log(cls, level, msg, args=None, exc_info=None, extra=None):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        # Add wrapping functionality here.
        if _srcfile:
            # IronPython doesn't track Python frames, so findCaller throws an
            # exception on some versions of IronPython. We trap it here so that
            # IronPython can use logging.
            try:
                fn, lno, func = cls.findCaller()
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        else:
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info:
            if not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()
        record = icon_logger.makeRecord(
            icon_logger.name, level, fn, lno, msg, args, exc_info, func, extra)
        icon_logger.handle(record)

    @classmethod
    def findCaller(cls):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        rv = "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _srcfile:
                f = f.f_back
                continue
            rv = (co.co_filename, f.f_lineno, co.co_name)
            break
        return rv
